#ifndef HARKIO_VECTOR3_HPP_
#define HARKIO_VECTOR3_HPP_

#include <boost/operators.hpp>

/** @file Vector3.hpp
 * @brief Vector3 class definition.
 */

namespace harkio2
{

/** @brief A orthogonal coordinate.
 * This class represents a three-dimensional orthogonal coordinate.
 */
template <typename T>
class Vector3 :
    private boost::equality_comparable<Vector3<T> > // for != operator
{
public:
    /** @brief Constructs an instance which represents the origin.
     */
    Vector3();

    /** @brief Constructs an instance with given values.
     * @param x the x-axis position
     * @param y the y-axis position
     * @param y the z-axis position
     */
    Vector3(const T& x, const T& y, const T& z);

    /// @brief Returns the x-axis position.
    const T& GetX() const { return x; }

    /// @brief Returns the y-axis position.
    const T& GetY() const { return y; }

    /// @brief Returns the z-axis position.
    const T& GetZ() const { return z; }

    /// @brief Checks whether this coordinate equals the given coordinate.
    bool operator ==(const Vector3& rhs) const;

private:
    /// @brief Axis positions.
    T x, y, z;
};

/** @brief Returns a normalized vector.
 * This function normalizes a given vector and returns it.
 * @param v A vector to normalize
 * @return A normalized vector
 */
template <typename T>
Vector3<T> Normalize(const Vector3<T>& v) throw ();


/* implementation */

template <typename T>
Vector3<T>::Vector3()
    : x(T()), y(T()), z(T())
{
}

template <typename T>
Vector3<T>::Vector3(const T& x, const T& y, const T& z)
    : x(x), y(y), z(z)
{
}

template <typename T>
bool Vector3<T>::operator ==(const Vector3<T>& rhs) const
{
    return x == rhs.x && y == rhs.y && z == rhs.z;
}

template <typename T>
Vector3<T> Normalize(const Vector3<T>& v) throw ()
{
    T x = v.GetX();
    T y = v.GetY();
    T z = v.GetZ();
    T l = sqrt(x*x + y*y + z*z);
    // Don't use zero vector, or zero-division will happen.
    return Vector3<T>(x/l, y/l, z/l);
}

}

#endif // HARKIO_VECTOR3_HPP_

