#define BOOST_TEST_MODULE "testcase for HarkFile class"
#define BOOST_TEST_DYN_LINK
#include "common.hpp"

#include <boost/test/unit_test.hpp>

#include <vector>
#include <stdexcept>
#include <boost/shared_ptr.hpp>
#include <harkio2/hark/DataSet.hpp>
#include <harkio2/hark/HarkFile.hpp>
#include <harkio2/hark/hgtf/Fmt.hpp>
#include <harkio2/hark/hgtf/Spec.hpp>
#include <harkio2/hark/hgtf/M2PG.hpp>
#include <harkio2/hark/hgtf/GTF.hpp>
#include <harkio2/hark/hgtf/SM.hpp>

using namespace boost;
using namespace std;

typedef boost::shared_ptr<harkio2::hark::DataSet> dsptr_t;

BOOST_AUTO_TEST_CASE(constructor_test)
{
    harkio2::hark::HarkFile h;
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 0);
}

BOOST_AUTO_TEST_CASE(add_test)
{
    harkio2::hark::HarkFile h;

    dsptr_t d(
            new harkio2::hark::hgtf::Fmt1_0());

    BOOST_REQUIRE(d);

    h.AddDataSet(d);

    BOOST_REQUIRE_EQUAL(h.GetDataSetsSize(), 1);
    BOOST_CHECK_THROW(h.GetDataSet(1), std::out_of_range);
    BOOST_REQUIRE_NO_THROW(h.GetDataSet(0));
    BOOST_CHECK_EQUAL(h.GetDataSet(0).get(), d.get());

    dsptr_t d2(
            new harkio2::hark::hgtf::Spec1_0());

    BOOST_REQUIRE(d2);

    h.AddDataSet(d);

    BOOST_REQUIRE_EQUAL(h.GetDataSetsSize(), 2);
}

BOOST_AUTO_TEST_CASE(get_test)
{
    harkio2::hark::HarkFile h;

    dsptr_t d0(new harkio2::hark::hgtf::Fmt1_0());
    dsptr_t d1(new harkio2::hark::hgtf::M2PG1_0());
    dsptr_t d2(new harkio2::hark::hgtf::Spec1_0());

    h.AddDataSet(d0);
    h.AddDataSet(d1);
    h.AddDataSet(d2);

    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);

    BOOST_REQUIRE(h.GetDataSet("hgtf::Fmt"));
    BOOST_CHECK_EQUAL(h.GetDataSet("hgtf::Fmt").get(), d0.get());
    boost::shared_ptr<harkio2::hark::hgtf::Fmt> d00 = dynamic_pointer_cast<harkio2::hark::hgtf::Fmt>(h.GetDataSet("hgtf::Fmt"));
    BOOST_CHECK(d00->GetVersion() == harkio2::Version(1, 0));

    BOOST_REQUIRE(h.GetDataSet("hgtf::M2PG"));
    BOOST_CHECK_EQUAL(h.GetDataSet("hgtf::M2PG").get(), d1.get());
    boost::shared_ptr<harkio2::hark::hgtf::M2PG> d10 = dynamic_pointer_cast<harkio2::hark::hgtf::M2PG>(h.GetDataSet("hgtf::M2PG"));
    BOOST_CHECK(d10->GetVersion() == harkio2::Version(1, 0));

    BOOST_REQUIRE(h.GetDataSet("hgtf::Spec"));
    BOOST_CHECK_EQUAL(h.GetDataSet("hgtf::Spec").get(), d2.get());
    boost::shared_ptr<harkio2::hark::hgtf::Spec> d20 = dynamic_pointer_cast<harkio2::hark::hgtf::Spec>(h.GetDataSet("hgtf::Spec"));
    BOOST_CHECK(d20->GetVersion() == harkio2::Version(1, 0));

    dsptr_t d3(new harkio2::hark::hgtf::Fmt1_0());
    h.AddDataSet(d3);

    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 4);
    BOOST_CHECK_EQUAL(h.GetDataSet("hgtf::Fmt"), d0);

    vector<dsptr_t> sets = h.SearchDataSets("hgtf::Fmt");
    BOOST_REQUIRE_EQUAL(sets.size(), 2);
    if (sets[0] == d0)
    {
        BOOST_CHECK_EQUAL(sets[1], d3);
    }
    else
    {
        BOOST_CHECK_EQUAL(sets[0], d3);
        BOOST_CHECK_EQUAL(sets[1], d0);
    }

}

BOOST_AUTO_TEST_CASE(remove_test)
{
    harkio2::hark::HarkFile h;

    dsptr_t d0(new harkio2::hark::hgtf::Fmt1_0());
    dsptr_t d1(new harkio2::hark::hgtf::M2PG1_0());
    dsptr_t d2(new harkio2::hark::hgtf::Spec1_0());

    h.AddDataSet(d0);
    h.AddDataSet(d1);
    h.AddDataSet(d2);

    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);

    h.RemoveDataSet(d1);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 2);
    BOOST_CHECK( ! h.GetDataSet("hgtf::M2PG"));

    h.AddDataSet(d1);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);
    BOOST_CHECK(h.GetDataSet("hgtf::M2PG"));

    h.AddDataSet(d0);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 4);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 2);

    h.RemoveDataSet(d0);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 1);

    h.RemoveDataSet(d0);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 2);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 0);
}

BOOST_AUTO_TEST_CASE(remove_test2)
{
    harkio2::hark::HarkFile h;

    dsptr_t d0(new harkio2::hark::hgtf::Fmt1_0());
    dsptr_t d1(new harkio2::hark::hgtf::M2PG1_0());
    dsptr_t d2(new harkio2::hark::hgtf::Spec1_0());

    h.AddDataSet(d0);
    h.AddDataSet(d1);
    h.AddDataSet(d2);

    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);

    h.RemoveDataSet("hgtf::M2PG");
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 2);
    BOOST_CHECK( ! h.GetDataSet("hgtf::M2PG"));

    h.AddDataSet(d1);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);
    BOOST_CHECK(h.GetDataSet("hgtf::M2PG"));

    h.AddDataSet(d0);
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 4);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 2);

    h.RemoveDataSet("hgtf::Fmt");
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 3);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 1);

    h.RemoveDataSet("hgtf::Fmt");
    BOOST_CHECK_EQUAL(h.GetDataSetsSize(), 2);
    BOOST_CHECK_EQUAL(h.SearchDataSets("hgtf::Fmt").size(), 0);
}

BOOST_AUTO_TEST_CASE(valid_test)
{
    harkio2::hark::HarkFile h;

    BOOST_CHECK(IsValidHarkFile(h));

    dsptr_t d0(new harkio2::hark::hgtf::Fmt1_0());
    dsptr_t d1(new harkio2::hark::hgtf::M2PG1_0());
    dsptr_t d2(new harkio2::hark::hgtf::Spec1_0());

    h.AddDataSet(d0);
    BOOST_CHECK(!IsValidHarkFile(h));

    h.AddDataSet(d1);
    BOOST_CHECK(IsValidHarkFile(h));

    h.AddDataSet(d2);
    BOOST_CHECK(IsValidHarkFile(h));

    h.RemoveDataSet(d0);
    BOOST_CHECK(!IsValidHarkFile(h));

}

