#define BOOST_TEST_MODULE "testcase for HarkConfigFileXMLWriter class"
#define BOOST_TEST_DYN_LINK
#include "common.hpp"

#include <vector>
#include <fstream>

#include <boost/test/unit_test.hpp>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

#include <harkio2/harkconfig/HarkConfigFile.hpp>
#include <harkio2/harkconfig/ImpulseResponseListFile.hpp>
#include <harkio2/harkconfig/TSPListFile.hpp>
#include <harkio2/harkconfig/MicArrayLocationFile.hpp>
#include <harkio2/harkconfig/NoiseLocationFile.hpp>
#include <harkio2/harkconfig/HarkConfigFileXMLWriter.hpp>
#include <harkio2/harkconfig/HarkConfigFileXMLReader.hpp>

BOOST_AUTO_TEST_CASE(write_impulseresponse_test)
{
    std::ofstream out("writetest_ir20.xml");
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> writer(new harkio2::harkconfig::HarkConfigFileXMLWriter(out));
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLReader> reader(new harkio2::harkconfig::HarkConfigFileXMLReader("writetest_ir20.xml"));
    BOOST_REQUIRE(writer != 0);
    BOOST_REQUIRE(reader != 0);
    boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile2_0> ir20(new harkio2::harkconfig::ImpulseResponseListFile2_0());
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFile> data;

    // make test data
    const unsigned int nmic = 2;
    const harkio2::FileType::Kind sourceFileType = harkio2::FileType::FLOAT;
    const Glib::ustring comment = u8R"(テストコメント)";

    ir20->SetNumMic(nmic);
    ir20->SetSourceFileType(sourceFileType);

    std::vector<harkio2::harkconfig::Source2> sources(2);
    std::vector<harkio2::harkconfig::File1> files;

    sources[0].SetPosition(harkio2::MakeDirectionFromVector<float>(harkio2::Vector3<float>(1, 2, 3)));
    files.resize(nmic);
    files[0].SetName(u8"日本語hoge0_ch0.flt");
    files[0].SetMicIndices({0});
    files[1].SetName(u8"hoge0_ch1.flt");
    files[1].SetMicIndices({1});
    sources[0].SetFiles(files);
    sources[0].SetCutStart(100);
    sources[0].SetCutEnd(200);
    sources[0].SetComment(u8"hoge0");

    sources[1].SetPosition(harkio2::MakeDirectionFromVector<float>(harkio2::Vector3<float>(4, 5, 6)));
    files.clear();
    files.resize(nmic);
    files[0].SetName(u8"hoge1_ch0.flt");
    files[0].SetMicIndices({0});
    files[1].SetName(u8"hoge1_ch1.flt");
    files[1].SetMicIndices({1});
    sources[1].SetFiles(files);
    sources[1].SetCutStart(300);
    sources[1].SetCutEnd(400);
    sources[1].SetComment(u8"hoge1");

    ir20->SetSources(sources);
    ir20->SetComment(comment);

    try
    {
        writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(ir20));
    }
    catch (const harkio2::InvalidHarkFile& e)
    {
        BOOST_ERROR((boost::format("InvalidHarkFile exception: %s") % e.what()).str());
    }

    out.close();

#if 1
    try
    {
        data = reader->Read();
    }
    catch (const harkio2::AnalyzeError& e)
    {
        BOOST_ERROR((boost::format("%d: %s") % e.line() % e.what()).str());
    }

    BOOST_REQUIRE(data != 0);

    BOOST_CHECK_EQUAL(data->GetId(), ir20->GetId());
    boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile> ir_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile>(data);
    BOOST_REQUIRE(ir_ != 0);

    BOOST_CHECK(ir_->GetVersion() == ir20->GetVersion());
    boost::shared_ptr<harkio2::harkconfig::ImpulseResponseListFile2_0> ir20_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::ImpulseResponseListFile2_0>(ir_);
    BOOST_REQUIRE(ir20_ != 0);

    BOOST_CHECK_EQUAL(ir20->GetNumMic(), ir20_->GetNumMic());
    BOOST_CHECK_EQUAL(ir20->GetSourceFileType(), ir20_->GetSourceFileType());
    BOOST_CHECK_EQUAL(ir20->GetComment(), ir20_->GetComment());

    BOOST_REQUIRE_EQUAL(ir20->GetSources().size(), ir20_->GetSources().size());
#endif

}

BOOST_AUTO_TEST_CASE(write_tsp_test)
{
    std::ofstream out("writetest_tsp20.xml");
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> writer(new harkio2::harkconfig::HarkConfigFileXMLWriter(out));
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLReader> reader(new harkio2::harkconfig::HarkConfigFileXMLReader("writetest_tsp20.xml"));
    BOOST_REQUIRE(writer != 0);
    BOOST_REQUIRE(reader != 0);
    boost::shared_ptr<harkio2::harkconfig::TSPListFile2_0> tsp20(new harkio2::harkconfig::TSPListFile2_0());
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFile> data;

    // make test data
    const unsigned int nmic = 2;
    const harkio2::FileType::Kind sourceFileType = harkio2::FileType::WAVE;
    const Glib::ustring comment = u8R"(テストコメント)";

    tsp20->SetSynchronousAverage(16);
    tsp20->SetOriginalImpulseFile(harkio2::TypedFilename(u8"TSPファイル.tsp", harkio2::FileType::FLOAT));
    tsp20->SetTSPOffset(16384);
    tsp20->SetTSPLength(1024);
    tsp20->SetSignalMax(32768);
    tsp20->SetNumMic(nmic);
    tsp20->SetSourceFileType(sourceFileType);

    std::vector<harkio2::harkconfig::Source2> sources(1);
    std::vector<harkio2::harkconfig::File1> files;

    sources[0].SetPosition(harkio2::MakeDirectionFromVector<float>(harkio2::Vector3<float>(1, 2, 3)));
    files.resize(1);
    files[0].SetName(u8"信号源ファイル.wav");
    files[0].SetMicIndices({0, 1});
    sources[0].SetFiles(files);
    sources[0].SetCutStart(1);
    sources[0].SetCutEnd(2);
    sources[0].SetComment(u8"TSPリストファイルコメント");

    tsp20->SetSources(sources);
    tsp20->SetComment(comment);

    try
    {
        writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(tsp20));
    }
    catch (const harkio2::InvalidHarkFile& e)
    {
        BOOST_ERROR((boost::format("InvalidHarkFile exception: %s") % e.what()).str());
    }

    out.close();

#if 1
    try
    {
        data = reader->Read();
    }
    catch (const harkio2::AnalyzeError& e)
    {
        BOOST_ERROR((boost::format("%d: %s") % e.line() % e.what()).str());
    }

    BOOST_REQUIRE(data != 0);

    BOOST_CHECK_EQUAL(data->GetId(), tsp20->GetId());
    boost::shared_ptr<harkio2::harkconfig::TSPListFile> tsp_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile>(data);
    BOOST_REQUIRE(tsp_ != 0);

    BOOST_CHECK(tsp_->GetVersion() == tsp20->GetVersion());
    boost::shared_ptr<harkio2::harkconfig::TSPListFile2_0> tsp20_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::TSPListFile2_0>(tsp_);
    BOOST_REQUIRE(tsp20_ != 0);

    BOOST_CHECK_EQUAL(tsp20->GetSynchronousAverage(), tsp20_->GetSynchronousAverage());
    BOOST_CHECK(tsp20->GetOriginalImpulseFile() == tsp20_->GetOriginalImpulseFile());
    BOOST_CHECK_EQUAL(tsp20->GetTSPOffset(), tsp20_->GetTSPOffset());
    BOOST_CHECK_EQUAL(tsp20->GetTSPLength(), tsp20_->GetTSPLength());
    BOOST_CHECK_EQUAL(tsp20->GetSignalMax(), tsp20_->GetSignalMax());
    BOOST_CHECK_EQUAL(tsp20->GetNumMic(), tsp20_->GetNumMic());
    BOOST_CHECK_EQUAL(tsp20->GetSourceFileType(), tsp20_->GetSourceFileType());
    BOOST_CHECK_EQUAL(tsp20->GetComment(), tsp20_->GetComment());

    BOOST_REQUIRE_EQUAL(tsp20->GetSources().size(), tsp20_->GetSources().size());
#endif

}

BOOST_AUTO_TEST_CASE(write_mic_test)
{
    std::ofstream out("writetest_mic20.xml");
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> writer(new harkio2::harkconfig::HarkConfigFileXMLWriter(out));
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLReader> reader(new harkio2::harkconfig::HarkConfigFileXMLReader("writetest_mic20.xml"));
    BOOST_REQUIRE(writer != 0);
    BOOST_REQUIRE(reader != 0);
    boost::shared_ptr<harkio2::harkconfig::MicArrayLocationFile2_0> mic20(new harkio2::harkconfig::MicArrayLocationFile2_0());
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFile> data;

    // make test data
    const unsigned int nmic = 2;
    const Glib::ustring comment = u8R"(テストコメント)";

    std::vector<harkio2::harkconfig::Mic1> mics(nmic);
    mics[0].SetMicIndex(0);
    mics[0].SetPosition(harkio2::MakePositionFromVector<float>(harkio2::Vector3<float>(0, 1, 0), 2));
    mics[0].SetComment(u8"マイク0");
    mics[1].SetMicIndex(1);
    mics[1].SetPosition(harkio2::MakePositionFromVector<float>(harkio2::Vector3<float>(1, 0, 0), 2));
    mics[1].SetComment(u8"マイク1");

    mic20->SetMics(mics);
    mic20->SetComment(comment);

    try
    {
        writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(mic20));
    }
    catch (const harkio2::InvalidHarkFile& e)
    {
        BOOST_ERROR((boost::format("InvalidHarkFile exception: %s") % e.what()).str());
    }

    out.close();

#if 1
    try
    {
        data = reader->Read();
    }
    catch (const harkio2::AnalyzeError& e)
    {
        BOOST_ERROR((boost::format("%d: %s") % e.line() % e.what()).str());
    }


    std::ofstream debugout("writetest_mic20_debug.xml");
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> debugwriter(new harkio2::harkconfig::HarkConfigFileXMLWriter(debugout));
    debugwriter->Write(data);


    BOOST_REQUIRE(data != 0);

    BOOST_CHECK_EQUAL(data->GetId(), mic20->GetId());
    boost::shared_ptr<harkio2::harkconfig::MicArrayLocationFile> mic_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::MicArrayLocationFile>(data);
    BOOST_REQUIRE(mic_ != 0);

    BOOST_CHECK(mic_->GetVersion() == mic20->GetVersion());
    boost::shared_ptr<harkio2::harkconfig::MicArrayLocationFile2_0> mic20_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::MicArrayLocationFile2_0>(mic_);
    BOOST_REQUIRE(mic20_ != 0);

    BOOST_CHECK_EQUAL(mic20->GetComment(), mic20_->GetComment());

    BOOST_REQUIRE_EQUAL(mic20->GetMics().size(), mic20_->GetMics().size());
    BOOST_CHECK_CLOSE(mic20->GetMics()[0].GetPosition().GetPositionVector().GetX(), mic20_->GetMics()[0].GetPosition().GetPositionVector().GetX(), 0.01);
    BOOST_CHECK_CLOSE(mic20->GetMics()[0].GetPosition().GetPositionVector().GetY(), mic20_->GetMics()[0].GetPosition().GetPositionVector().GetY(), 0.01);
    BOOST_CHECK_CLOSE(mic20->GetMics()[1].GetPosition().GetPositionVector().GetX(), mic20_->GetMics()[1].GetPosition().GetPositionVector().GetX(), 0.01);
    BOOST_CHECK_CLOSE(mic20->GetMics()[1].GetPosition().GetPositionVector().GetY(), mic20_->GetMics()[1].GetPosition().GetPositionVector().GetY(), 0.01);
#endif

}

BOOST_AUTO_TEST_CASE(write_noise_test)
{
    std::ofstream out("writetest_noise20.xml");
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> writer(new harkio2::harkconfig::HarkConfigFileXMLWriter(out));
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLReader> reader(new harkio2::harkconfig::HarkConfigFileXMLReader("writetest_noise20.xml"));
    BOOST_REQUIRE(writer != 0);
    BOOST_REQUIRE(reader != 0);
    boost::shared_ptr<harkio2::harkconfig::NoiseLocationFile2_0> noise20(new harkio2::harkconfig::NoiseLocationFile2_0());
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFile> data;

    // make test data
    const Glib::ustring comment = u8R"(テストコメント)";

    std::vector<harkio2::harkconfig::NoiseSource1> noises(1);
    noises[0].SetNoiseIndex(0);
    noises[0].SetPosition(harkio2::MakeDirectionFromVector<float>(harkio2::Vector3<float>(1, 2, 3)));
    noises[0].SetComment(u8"ノイズ0");

    noise20->SetNoiseSources(noises);
    noise20->SetComment(comment);

    try
    {
        writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(noise20));
    }
    catch (const harkio2::InvalidHarkFile& e)
    {
        BOOST_ERROR((boost::format("InvalidHarkFile exception: %s") % e.what()).str());
    }

    out.close();

#if 1
    try
    {
        data = reader->Read();
    }
    catch (const harkio2::AnalyzeError& e)
    {
        BOOST_ERROR((boost::format("%d: %s") % e.line() % e.what()).str());
    }

    BOOST_REQUIRE(data != 0);

    BOOST_CHECK_EQUAL(data->GetId(), noise20->GetId());
    boost::shared_ptr<harkio2::harkconfig::NoiseLocationFile> noise_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::NoiseLocationFile>(data);
    BOOST_REQUIRE(noise_ != 0);

    BOOST_CHECK(noise_->GetVersion() == noise20->GetVersion());
    boost::shared_ptr<harkio2::harkconfig::NoiseLocationFile2_0> noise20_ =
        boost::dynamic_pointer_cast<harkio2::harkconfig::NoiseLocationFile2_0>(noise_);
    BOOST_REQUIRE(noise20_ != 0);

    BOOST_CHECK_EQUAL(noise20->GetComment(), noise20_->GetComment());

    BOOST_REQUIRE_EQUAL(noise20->GetNoiseSources().size(), noise20_->GetNoiseSources().size());
    BOOST_CHECK_CLOSE(noise20->GetNoiseSources()[0].GetPosition().GetUnitVector().GetX(), noise20_->GetNoiseSources()[0].GetPosition().GetUnitVector().GetX(), 0.01);
    BOOST_CHECK_CLOSE(noise20->GetNoiseSources()[0].GetPosition().GetUnitVector().GetY(), noise20_->GetNoiseSources()[0].GetPosition().GetUnitVector().GetY(), 0.01);
#endif

}

BOOST_AUTO_TEST_CASE(invalid_nummic_test)
{
    std::stringstream out;
    boost::shared_ptr<harkio2::harkconfig::HarkConfigFileXMLWriter> writer(new harkio2::harkconfig::HarkConfigFileXMLWriter(out));
    BOOST_REQUIRE(writer != 0);
    boost::shared_ptr<harkio2::harkconfig::TSPListFile2_0> tsp20(new harkio2::harkconfig::TSPListFile2_0());

    // make test data
    const unsigned int nmic = 2;
    const harkio2::FileType::Kind sourceFileType = harkio2::FileType::WAVE;
    const Glib::ustring comment = u8R"(テストコメント)";

    tsp20->SetSynchronousAverage(16);
    tsp20->SetOriginalImpulseFile(harkio2::TypedFilename(u8"TSPファイル.tsp", harkio2::FileType::FLOAT));
    tsp20->SetTSPOffset(16384);
    tsp20->SetTSPLength(1024);
    tsp20->SetSignalMax(32768);
    //tsp20->SetNumMic(nmic);
    tsp20->SetSourceFileType(sourceFileType);

    std::vector<harkio2::harkconfig::Source2> sources(1);
    std::vector<harkio2::harkconfig::File1> files;

    sources[0].SetPosition(harkio2::MakeDirectionFromVector<float>(harkio2::Vector3<float>(1, 2, 3)));
    files.resize(1);
    files[0].SetName(u8"信号源ファイル.wav");
    files[0].SetMicIndices({0, 1});
    sources[0].SetFiles(files);
    sources[0].SetCutStart(1);
    sources[0].SetCutEnd(2);
    sources[0].SetComment(u8"TSPリストファイルコメント");

    tsp20->SetSources(sources);
    tsp20->SetComment(comment);

    BOOST_CHECK_NO_THROW(
            writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(tsp20)));

    out.str("");
    tsp20->SetNumMic(nmic + 1); // invalid data
    BOOST_CHECK_THROW(
            writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(tsp20)),
            harkio2::InvalidHarkFile);

    out.str("");
    tsp20->SetNumMic(nmic); // valid data
    BOOST_CHECK_NO_THROW(
            writer->Write(boost::dynamic_pointer_cast<harkio2::harkconfig::HarkConfigFile>(tsp20)));

}

